"""The Lorenz attractor is a set of chaotic solutions of
a particular system of ordinary differential equations"""
from vedo import *

dt = 0.002
y = (25.0, -10.0, -7.0)  # Starting point (initial condition)
pts, vel = [], []

for t in np.linspace(0, 20, int(20 / dt)):
    # Integrate a funny differential equation
    dydt = np.array(
        [-8 / 3.0 * y[0] + y[1] * y[2],
         -10.0 * (y[1] - y[2]),
         -y[1] * y[0] + 28.0 * y[1] - y[2]]
    )
    y = y + dydt * dt
    v = np.clip([np.linalg.norm(dydt) * 0.005], 0, 1)[0]  # color by speed
    vel.append(v)
    pts.append(y)

plt  = Plotter(title="Lorenz attractor", axes=dict(yzGrid=True))
plt += Line(pts).lw(2).cmap("brg", vel)
plt += __doc__
plt.show(viewup='z').close()



from vedo import dataurl, printc, precision
from vedo import Plotter, Picture, Spline, Points, Text2D
# from vedo.applications import SplinePlotter  # ready to use class!

#########################################################################
class SplinePlotter(Plotter):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.cpoints = []
        self.points = None
        self.spline = None

    def onLeftClick(self, evt):
        if not evt.actor: return
        p = evt.picked3d + [0,0,1]
        self.cpoints.append(p)
        self.update()
        printc("Added point:", precision(p[:2],4), c='g')

    def onRightClick(self, evt):
        if evt.actor and len(self.cpoints)>0:
            self.cpoints.pop() # pop removes from the list the last pt
            self.update()
            printc("Deleted last point", c="r")

    def update(self):
        self.remove([self.spline, self.points])  # remove old points and spline
        self.points = Points(self.cpoints).ps(10).c('purple5')
        self.points.pickable(False)  # avoid picking the same point
        if len(self.cpoints) > 2:
            self.spline = Spline(self.cpoints, closed=False).c('yellow5').lw(3)
            self.add(self.points, self.spline)
        else:
            self.add(self.points)

    def keyPress(self, evt):
        if evt.keyPressed == 'c':
            self.cpoints = []
            self.remove(self.spline, self.points).render()
            printc("==== Cleared all points ====", c="r", invert=True)


##############################################################################
if __name__ == "__main__":

    filename = dataurl+"images/Mouse-_embryo_E11.5.jpg"
    pic = Picture(filename)

    t = """Click to add a point
    Right-click to remove it
    Drag mouse to change constrast
    Press c to clear points"""
    instrucs = Text2D(t, pos='bottom-left', c='white', bg='green', font='Quikhand')

    plt = SplinePlotter(axes=True, bg='blackboard')
    plt.addCallback('KeyPress', plt.keyPress)
    plt.addCallback('LeftButtonPress', plt.onLeftClick)
    plt.addCallback('RightButtonPress', plt.onRightClick)
    plt.show(filename, pic, instrucs, mode='image', zoom=1.2)
    plt.close()





import vtk
from vedo import Grid, Tensors, show

domain = Grid(res=[5,5], c='gray')

# Generate random attributes on a plane
ag = vtk.vtkRandomAttributeGenerator()
ag.SetInputData(domain.polydata())
ag.GenerateAllDataOn()
ag.Update()

ts = Tensors(ag.GetOutput(), scale=0.1)
ts.print()

show(domain, ts).close()




"""Draw the shadows and trailing lines of 2 planes. Not really
a simulation.. just a way to illustrate how to move objects around!"""
from vedo import *

settings.allowInteraction = True # if ESC button is hit during the loop

world = Box([0,0,0], 30, 16, 8).wireframe()

plane1 = Mesh(dataurl+"cessna.vtk").c("green")
plane1.pos(-15, 2, 0.14).addTrail(n=200)
plane1.addShadow('z', -4).addShadow('y', 8)

plane2 = plane1.clone().c("tomato")
plane2.pos(-15,-2,-0.21).addTrail(n=200)
plane2.addShadow('z', -4).addShadow('y', 8)

# Setup the scene
plt = Plotter(axes=1, interactive=False)
plt.show(world, plane1, plane2, __doc__, viewup="z")

for t in np.arange(0, 3.2, 0.02):
    plane1.pos(9*t-15, 2-t, sin(3-t)).rotateX(0+t) # make up some movement
    plane2.pos(8*t-15, t-2, sin(t-3)).rotateX(2-t) # for the 2 planes

    plt.show(world, plane1, plane2)

    if plt.escaped:
        break  # if ESC button is hit during the loop

plt.interactive().close()



"""Take 2 clouds of points, source and target, and morph
the plane using thin plate splines as a model.
The fitting minimizes the distance to a subset of the target cloud"""
from vedo import printc, Points, Grid, Arrows, Lines, Plotter
import scipy.optimize as opt
import numpy as np
np.random.seed(2)

class Morpher(Plotter):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.source = None
        self.morphed_source = None
        self.target = None
        self.bound = None
        self.sigma = 1  # stiffness of the mesh
        self.method = "SLSQP"  # 'SLSQP', 'L-BFGS-B', 'TNC' ...
        self.fitTolerance = 1e-6
        self.fitResult = None
        self.chi2 = 1.0e30
        self.npts = None
        self.ptsource = []
        self.pttarget = []


    def _func(self, pars):
        shift = np.array(np.split(pars,2)).T # recreate the shift vectors
        z = np.zeros((self.npts,1))
        shift = np.append(shift, z, axis=1) # make them 3d
        self.morphed_source = self.source.clone().warp(self.ptsource,
                                                       self.ptsource + shift,
                                                       sigma=self.sigma,
                                                       mode="2d")
        d = self.morphed_source.points() - self.target.points()
        chi2 = np.sum(np.multiply(d,d))#/len(d)
        if chi2 < self.chi2:
            printc("new minimum ->", chi2)
            self.chi2 = chi2
        return chi2

    # ------------------------------------------------------- Fit
    def morph(self):
        print("\n..minimizing with " + self.method)
        self.morphed_source = self.source.clone()

        self.ptsource = self.source.points()[:self.npts] # pick the first npts points
        self.pttarget = self.target.points()[:self.npts]

        delta = self.pttarget - self.ptsource
        x0 = delta[:,(0,1)].T.ravel() # initial guess, a flat list of x and y shifts
        bnds = [(-self.bound, self.bound)] * (2*self.npts)
        res = opt.minimize(self._func, x0, bounds=bnds, method=self.method, tol=self.fitTolerance)
        self.fitResult = res
        # recalculate the last step:
        self._func(res["x"])

    # ------------------------------------------------------- Visualization
    def draw_shapes(self):
        sb = self.source.bounds()
        x1, x2, y1, y2, z1, z2 = sb
        maxb = max(x2-x1, y2-y1)
        grid0 = Grid(self.source.centerOfMass(), s=[maxb,maxb], res=[40,40])
        T = self.morphed_source.getTransform()
        grid1 = grid0.alpha(0.3).wireframe(0).clone().applyTransform(T) # warp the grid
        arrows = Arrows(self.ptsource, self.pttarget, alpha=0.5, s=3).c("k")
        lines = Lines(self.source, self.target).c('db')
        mlines = Lines(self.morphed_source, self.target).c('db')

        self.at(0).show(grid0, self.source, self.target, lines, arrows, __doc__)
        self.at(1).show(grid1, self.morphed_source, self.target, mlines,
                        f"morphed source (green) vs target (red)\nNDF = {2*self.npts}")


#################################
if __name__ == "__main__":

    # make up a source random cloud
    pts_s = np.random.randn(25, 2)
    pts_t = pts_s + np.sin(2*pts_s)/5 # and distort it

    mr = Morpher(N=2)
    mr.source = Points(pts_s, r=20, c="g", alpha=0.5)
    mr.target = Points(pts_t, r=10, c="r", alpha=1.0)

    mr.bound = 2  # limits the x and y shift
    mr.npts  = 6  # allow move only a subset of points (implicitly sets the NDF of the fit)
    mr.sigma = 1. # stiffness of the mesh (1=max stiffness)

    mr.morph()

    #now mr.msource contains the modified/morphed source.
    mr.draw_shapes()
    mr.interactive().close()



"""Earthquakes of magnitude 2.5+ in the past 30 days
areas are proportional to energy release
[hover mouse to get more info]"""
import pandas
import numpy as np
from vedo import download, Picture, ProgressBar, colorMap, Plotter, Text2D, GeoCircle

num = 50  # nr of earthquakes to be visualized to define a time window
path = download("https://earthquake.usgs.gov/earthquakes/feed/v1.0/summary/2.5_month.csv")
usecols = ['time','place','latitude','longitude','depth','mag']
data = pandas.read_csv(path, usecols=usecols)[usecols][::-1].reset_index(drop=True) # reverse list

pic = Picture("https://eoimages.gsfc.nasa.gov/images/imagerecords/147000/147190/eo_base_2020_clean_3600x1800.png")
pic.pickable(False).level(185).window(120)  # add some contrast to the original image
scale = [pic.shape[0]/2, pic.shape[1]/2, 1]

centers = []
pb = ProgressBar(0, len(data))
for i, d in data.iterrows():
    pb.print("Parsing USGS data..")
    M = d['mag']                                       # earthquake estimated magnitude
    E = np.sqrt(np.exp(5.24+1.44*M) * scale[0])/10000  # empirical formula for sqrt(energy_release(M))
    rgb = colorMap(E, name='Reds', vmin=0, vmax=7)     # map energy to color
    lat, lon = np.deg2rad(d['latitude']), np.deg2rad(d['longitude'])
    ce = GeoCircle(lat, lon, E/50).scale(scale).z(num/M).c(rgb).lw(0.1).useBounds(False)
    ce.time = i
    ce.info = '\n'.join(str(d).split('\n')[:-1])       # remove of the last line in string d
    if i < len(data)-num: ce.off()                     # switch off older ones: make circles invisible
    centers.append(ce)


def sliderfunc(widget, event):
    value = widget.GetRepresentation().GetValue()      # get the slider current value
    widget.GetRepresentation().SetTitleText(f"{data['time'][int(value)][:10]}")
    for ce in centers:
        isinside = abs(value-ce.time) < num            # switch on if inside of time window
        ce.on() if isinside else ce.off()
    plt.render()

plt = Plotter(size=(2200,1100), title="Earthquake Browser")
plt.addSlider2D(sliderfunc, 0, len(centers)-1, value=len(centers)-1, showValue=False, title="today")
plt.addHoverLegend(useInfo=True, alpha=1, c='white', bg='red2', s=1)
comment = Text2D(__doc__, bg='green9', alpha=0.7, font='Ubuntu')
plt.show(pic, centers, comment, zoom=2.27, mode='image').close()



"""Interpolate a vectorial field using:

Thin Plate Spline or Radial Basis Function.

Example shows how to share the same vtkCamera
between different Plotter windows.
"""
from vedo import Plotter, Points, Arrows, show
import numpy as np


ls = np.linspace(0, 10, 8)
X, Y, Z = np.meshgrid(ls, ls, ls)
xr, yr, zr = X.ravel(), Y.ravel(), Z.ravel()
positions = np.vstack([xr, yr, zr])

sources = [(5, 8, 5), (8, 5, 5), (5, 2, 5)]
deltas = [(1, 1, 0.2), (1, 0, -0.8), (1, -1, 0.2)]

apos = Points(positions, r=2)

# for p in apos.points(): ####### Uncomment to fix some points.
#    if abs(p[2]-5) > 4.999:  # differences btw RBF and thinplate
#        sources.append(p)    # will become much smaller.
#        deltas.append(np.zeros(3))
sources = np.array(sources)
deltas = np.array(deltas)

src = Points(sources, c="r", r=12)
trs = Points(sources + deltas, c="v", r=12)
arr = Arrows(sources, sources + deltas)

################################################# warp using Thin Plate Splines
warped = apos.clone().warp(sources, sources+deltas)
warped.alpha(0.4).color("lg").pointSize(10)
allarr = Arrows(apos.points(), warped.points())

set1 = [apos, warped, src, trs, arr, __doc__]
plt1 = show([set1, allarr], N=2, bg='bb')  # returns the Plotter class


################################################# RBF
from scipy.interpolate import Rbf

x, y, z = sources[:, 0], sources[:, 1], sources[:, 2]
dx, dy, dz = deltas[:, 0], deltas[:, 1], deltas[:, 2]

itrx = Rbf(x, y, z, dx)  # Radial Basis Function interpolator:
itry = Rbf(x, y, z, dy)  #  interoplate the deltas in each separate
itrz = Rbf(x, y, z, dz)  #  cartesian dimension

positions_x = itrx(xr, yr, zr) + xr
positions_y = itry(xr, yr, zr) + yr
positions_z = itrz(xr, yr, zr) + zr
positions_rbf = np.vstack([positions_x, positions_y, positions_z])

warped_rbf = Points(positions_rbf, r=2).alpha(0.4).color("lg").pointSize(10)
allarr_rbf = Arrows(apos.points(), warped_rbf.points())

arr = Arrows(sources, sources + deltas)

plt2 = Plotter(N=2, pos=(200, 300), bg='bb')
plt2.camera = plt1.camera  # share the same camera with previous Plotter

plt2.at(0).show("Radial Basis Function", apos, warped_rbf, src, trs, arr)
plt2.at(1).show(allarr_rbf)

plt2.interactive().close()
plt1.close()

